;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "benchmark-init" "20260108.1447"
  "Benchmarks for require and load calls."
  '((emacs "24.4"))
  :url "https://github.com/dholm/benchmark-init-el"
  :commit "54b9703389f25012e4cc20fe4a0d4ea253ce4820"
  :revdesc "54b9703389f2"
  :keywords '("convenience" "benchmark")
  :maintainers '(("David Holm" . "dholmster@gmail.com")))
